import asyncio
import types
from .base import _wrap, Foo, Const, Share  # noqa: F401
from .calculations import *  # noqa: F401, F403
from .utils import *  # noqa: F401, F403
from .input import *  # noqa: F401, F403
from .output import *  # noqa: F401, F403


async def _run(foo, **kwargs):
    foo = _wrap(foo, kwargs)
    ret = []
    try:
        async for item in foo():
            if isinstance(item, types.AsyncGeneratorType):
                async for i in item:
                    ret.append(i)
            elif isinstance(item, types.CoroutineType):
                ret.append(await item)
            else:
                ret.append(item)

    except KeyboardInterrupt:
        print('Terminating...')
    return ret


def run(foo, **kwargs):
    loop = asyncio.get_event_loop()
    x = loop.run_until_complete(_run(foo, **kwargs))
    return x



from ujson import loads as load_json
from websocket import create_connection
from ..base import _wrap
from ...base import StreamNone, StreamEnd
from ...thread import run


def WebSocket(url, *args, **kwargs):
    return AsyncWebSocket(url, *args, **kwargs)


def AsyncWebSocket(url, json=False, wrap=False):
    async def _listen(url, json, wrap):
        ws = create_connection(url)
        for x in run(ws.recv):
            if isinstance(x, StreamNone):
                continue
            elif not x or isinstance(x, StreamEnd):
                break

            if json:
                x = load_json(x)
            if wrap:
                x = [x]
            yield x

    return _wrap(_listen, dict(url=url, json=json, wrap=wrap), name='WebSocket')



import time
import requests
from ..base import _wrap


def HTTP(url, *args, **kwargs):
    return AsyncHTTP(url, *args, **kwargs)


def AsyncHTTP(url, interval=1, repeat=1, json=False, wrap=False, field=None, proxies=None, cookies=None):
    async def _req(url, interval=1, repeat=1, json=False, wrap=False, field=None, proxies=None, cookies=None):
        count = 0
        while count < repeat:
            msg = requests.get(url, cookies=cookies, proxies=proxies)

            if msg is None or msg.status_code != 200:
                break

            if json:
                msg = msg.json()

            if field:
                msg = msg[field]

            if wrap:
                msg = [msg]

            yield msg

            if interval:
                time.sleep(interval)
            if repeat >= 0:
                count += 1

    return _wrap(_req, dict(url=url, interval=interval, repeat=repeat, json=json, wrap=wrap, field=field, proxies=proxies, cookies=cookies), name='HTTP')




def split(data, callback1, callback2):
    return (callback1(data), callback2(data))


def map(data, *callbacks):
    return (callback(data) for callback in callbacks)


def merge(data1, data2, callback):
    return callback((data1, data2))


def reduce(callback, *datas):
    return callback((data for data in datas))



import requests
import time
from confluent_kafka import Consumer, KafkaError
from ujson import loads as load_json
from websocket import create_connection
from socketIO_client_nexus import SocketIO as SIO
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


from ..base import StreamNone, StreamEnd
from ..thread import run


def ws(url, callback, json=False, wrap=False):
    ws = create_connection(url)
    for x in run(ws.recv):
        if isinstance(x, StreamNone):
            continue
        elif not x or isinstance(x, StreamEnd):
            break

        if json:
            x = load_json(x)
        if wrap:
            x = [x]
        callback(x)


def http(url, callback, interval=1, repeat=1, json=False, wrap=False, field=None, proxies=None, cookies=None):
    count = 0
    while count < repeat:
        msg = requests.get(url, cookies=cookies, proxies=proxies)

        if msg is None or msg.status_code != 200:
            break

        if json:
            msg = msg.json()

        if field:
            msg = msg[field]

        if wrap:
            msg = [msg]

        callback(msg)

        if interval:
            time.sleep(interval)
        if repeat >= 0:
            count += 1


def socketio(url, callback, channel='', field='', sendinit=None, json=False, wrap=False, interval=1):
    o = urlparse(url)
    socketIO = SIO(o.scheme + '://' + o.netloc, o.port)
    if sendinit:
        socketIO.emit(sendinit)

    while True:
        _data = []
        socketIO.on(channel, lambda data: _data.append(data))
        socketIO.wait(seconds=interval)
        for msg in _data:
            if json:
                msg = json.loads(msg)

            if field:
                msg = msg[field]

            if wrap:
                msg = [msg]

            callback(msg)


def kafka(callback, servers, group, topics, json=False, wrap=False, interval=1):
    c = Consumer({
        'bootstrap.servers': servers,
        'group.id': group,
        'default.topic.config': {
            'auto.offset.reset': 'smallest'
        }
    })

    if not isinstance(topics, list):
        topics = [topics]
    c.subscribe(topics)

    def _listen(consumer, json, wrap, interval):
        while True:
            msg = consumer.poll(interval)

            if msg is None:
                continue
            if msg.error():
                if msg.error().code() == KafkaError._PARTITION_EOF:
                    continue
                else:
                    break

            msg = msg.value().decode('utf-8')

            if not msg:
                break
            if json:
                msg = load_json(msg)
            if wrap:
                msg = [msg]
            callback(msg)



import types
from pprint import pprint
from ..base import _wrap, FunctionWrapper


def Print(foo, foo_kwargs=None):
    foo_kwargs = foo_kwargs or {}
    foo = _wrap(foo, foo_kwargs)

    async def _print(foo):
        async for r in foo():
            if isinstance(r, types.AsyncGeneratorType):
                async for x in r:
                    print(x)
                    yield x
            elif isinstance(r, types.CoroutineType):
                print(r)
                yield await r
            else:
                print(r)
                yield r

    return _wrap(_print, dict(foo=foo), name='Print', wraps=(foo,), share=foo)


def Graph(f_wrap):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('ViewGraph expects tributary')
    return f_wrap.view(0)[0]


def PPrint(f_wrap):
    pprint(Graph(f_wrap))


def GraphViz(f_wrap, name='Graph'):
    d = Graph(f_wrap)

    from graphviz import Digraph
    dot = Digraph(name, strict=True)
    dot.format = 'png'

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                dot.node(d)
                dot.edge(d, parent)

            else:
                for k in d:
                    dot.node(k)
                    rec(d[k], k)
                    dot.edge(k, parent)

    for k in d:
        dot.node(k)
        rec(d[k], k)

    return dot



import ujson
from confluent_kafka import Producer
from ..base import _wrap


def Kafka(foo, foo_kwargs=None, **kafka_kwargs):
    return AsyncKafka(foo, foo_kwargs, **kafka_kwargs)


def AsyncKafka(foo, foo_kwargs=None, servers='', topic='', json=False, wrap=False):
    foo = _wrap(foo, foo_kwargs or {})

    p = Producer({'bootstrap.servers': servers})

    async def _send(foo, producer, topic, json, wrap):
        ret = []
        async for data in foo():
            # Trigger any available delivery report callbacks from previous produce() calls
            producer.poll(0)

            if wrap:
                data = [data]

            if json:
                data = ujson.dumps(data)

            producer.produce(topic, data.encode('utf-8'), callback=lambda *args: ret.append(args))

            for data in ret:
                yield data
            ret = []

    return _wrap(_send, dict(foo=foo, producer=p, topic=topic, json=json, wrap=wrap), name='Kafka')



from socketIO_client_nexus import SocketIO as SIO
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
from ..base import _wrap


def SocketIO(url, *args, **kwargs):
    return SyncSocketIO(url, *args, **kwargs)


def SyncSocketIO(url, channel='', field='', sendinit=None, json=False, wrap=False, interval=1):
    o = urlparse(url)
    socketIO = SIO(o.scheme + '://' + o.netloc, o.port)
    if sendinit:
        socketIO.emit(sendinit)

    def _sio(url, channel, field='', json=False, wrap=False, interval=1):
        while True:
            _data = []
            socketIO.on(channel, lambda data: _data.append(data))
            socketIO.wait(seconds=interval)
            for msg in _data:
                # FIXME clear _data
                if json:
                    msg = json.loads(msg)

                if field:
                    msg = msg[field]

                if wrap:
                    msg = [msg]

                yield msg

    return _wrap(_sio, dict(url=url, channel=channel, field=field, json=json, wrap=wrap, interval=interval), name='SocketIO')



from pprint import pprint
from functools import partial

from ..base import _wrap, FunctionWrapper
from .kafka import Kafka as KafkaSink  # noqa: F401
from .http import HTTP as HTTPSink  # noqa: F401


def Print(foo, foo_kwargs=None):
    foo = _wrap(foo, foo_kwargs or {})

    def _print(foo):
        for r in foo():
            print(r)

    return _wrap(_print, dict(foo=foo), name='Print', wraps=(foo,), share=foo)


def Graph(f_wrap):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('ViewGraph expects tributary')
    return f_wrap.view(0)[0]


def PPrint(f_wrap):
    pprint(Graph(f_wrap))


def GraphViz(f_wrap, name='Graph'):
    d = Graph(f_wrap)
    from graphviz import Digraph
    dot = Digraph(name, strict=True)
    dot.format = 'png'

    def rec(nodes, parent):
        for d in nodes:
            if not isinstance(d, dict):
                dot.node(d)
                dot.edge(d, parent)

            else:
                for k in d:
                    dot.node(k)
                    rec(d[k], k)
                    dot.edge(k, parent)

    for k in d:
        dot.node(k)
        rec(d[k], k)

    return dot


def Perspective(foo, foo_kwargs=None, **psp_kwargs):
    foo = _wrap(foo, foo_kwargs or {})

    from perspective import PerspectiveWidget
    p = PerspectiveWidget([], **psp_kwargs)

    def _perspective(foo):
        for r in foo():
            p.update(r)
            yield r

    from IPython.display import display
    display(p)

    return _wrap(_perspective, dict(foo=foo), name='Perspective', wraps=(foo,), share=foo)


def FunctionalSink(f_wrap, callback, callback_kwargs):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('Functional expects tributary')

    callback = partial(callback, **callback_kwargs)

    def _foo(foo, callback_):
        for x in foo():
            yield callback(foo)

    return _wrap(_foo, dict(foo=f_wrap), name='Functional', wraps=(f_wrap,))



import time
import types
from .base import _wrap, FunctionWrapper, Foo, Const
try:
    from itertools import izip as zip
except ImportError:
    pass


def Timer(foo_or_val, kwargs=None, interval=1, repeat=0):
    if not isinstance(foo_or_val, types.FunctionType):
        foo = Const(foo_or_val)
    else:
        foo = Foo(foo_or_val, kwargs or {})

    def _repeater(foo, repeat, interval):
        while repeat > 0:
            t1 = time.time()
            yield foo()
            t2 = time.time()

            if interval > 0:
                # sleep for rest of time that _p didnt take
                time.sleep(max(0, interval-(t2-t1)))
            repeat -= 1

    return _wrap(_repeater, dict(foo=foo, repeat=repeat, interval=interval), name='Timer', wraps=(foo,), share=foo)


def Delay(f_wrap, delay=1):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('Delay expects a tributary')

    def _delay(f_wrap, delay):
        for f in f_wrap():
            yield f
            time.sleep(delay)

    return _wrap(_delay, dict(f_wrap=f_wrap, delay=delay), name='Delay', wraps=(f_wrap,), share=f_wrap)


def State(foo, foo_kwargs=None, **state):
    foo = _wrap(foo, foo_kwargs or {}, name=foo.__name__, wraps=(foo,), state=state)
    return foo


def Apply(foo, f_wrap, foo_kwargs=None):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('Apply expects a tributary')

    foo = Foo(foo, foo_kwargs or {})
    foo._wraps = foo._wraps + (f_wrap, )

    def _apply(foo):
        for f in f_wrap():
            yield foo(f)

    return _wrap(_apply, dict(foo=foo), name='Apply', wraps=(foo,), share=foo)


def Window(foo, foo_kwargs=None, size=-1, full_only=True):
    foo = Foo(foo, foo_kwargs or {})

    accum = []

    def _window(foo, size, full_only, accum):
        for x in foo():
            if size == 0:
                yield x
            else:
                accum.append(x)

                if size > 0:
                    accum = accum[-size:]
                if full_only:
                    if len(accum) == size or size == -1:
                        yield accum
                else:
                    yield accum

    return _wrap(_window, dict(foo=foo, size=size, full_only=full_only, accum=accum), name='Window', wraps=(foo,), share=foo)


def Unroll(foo, foo_kwargs):
    foo = Foo(foo, foo_kwargs or {})

    def _unroll(foo):
        ret = foo()
        if isinstance(ret, list) or isinstance(ret, types.GeneratorType):
            for f in ret:
                yield f

    return _wrap(_unroll, dict(foo=foo), name='Unroll', wraps=(foo,), share=foo)


def UnrollDataFrame(df, json=True, wrap=False):
    def _unrolldf(val):
        for i in range(len(df)):
            row = df.iloc[i]
            if json:
                data = row.to_dict()
                data['index'] = row.name
                yield data
            else:
                yield row

    return _wrap(_unrolldf, dict(val=df), name='UnrollDataFrame', wraps=(df,))


def Merge(f_wrap1, f_wrap2):
    if not isinstance(f_wrap1, FunctionWrapper):
        raise Exception('Merge expects a tributary')

    if not isinstance(f_wrap2, FunctionWrapper):
        raise Exception('Merge expects a tributary')

    def _merge(foo1, foo2):
        for gen1, gen2 in zip(foo1(), foo2()):
            if isinstance(gen1, types.GeneratorType) and \
               isinstance(gen2, types.GeneratorType):
                for f1, f2 in zip(gen1, gen2):
                    yield [f1, f2]
            elif isinstance(gen1, types.GeneratorType):
                for f1 in gen1:
                    yield [f1, gen2]
            elif isinstance(gen2, types.GeneratorType):
                for f2 in gen2:
                    yield [gen1, f2]
            else:
                yield [gen1, gen2]

    return _wrap(_merge, dict(foo1=f_wrap1, foo2=f_wrap2), name='Merge', wraps=(f_wrap1, f_wrap2), share=None)


def ListMerge(f_wrap1, f_wrap2):
    if not isinstance(f_wrap1, FunctionWrapper):
        raise Exception('Merge expects a tributary')

    if not isinstance(f_wrap2, FunctionWrapper):
        raise Exception('Merge expects a tributary')

    def _merge(foo1, foo2):
        for gen1, gen2 in zip(foo1(), foo2()):
            if isinstance(gen1, types.GeneratorType) and \
               isinstance(gen2, types.GeneratorType):
                for f1, f2 in zip(gen1, gen2):
                    ret = []
                    ret.extend(f1)
                    ret.extend(f1)
                    yield ret
            elif isinstance(gen1, types.GeneratorType):
                for f1 in gen1:
                    ret = []
                    ret.extend(f1)
                    ret.extend(gen2)
                    yield ret
            elif isinstance(gen2, types.GeneratorType):
                for f2 in gen2:
                    ret = []
                    ret.extend(gen1)
                    ret.extend(f2)
                    yield ret
            else:
                ret = []
                ret.extend(gen1)
                ret.extend(gen2)
                yield ret

    return _wrap(_merge, dict(foo1=f_wrap1, foo2=f_wrap2), name='ListMerge', wraps=(f_wrap1, f_wrap2), share=None)


def DictMerge(f_wrap1, f_wrap2):
    if not isinstance(f_wrap1, FunctionWrapper):
        raise Exception('Merge expects a tributary')

    if not isinstance(f_wrap2, FunctionWrapper):
        raise Exception('Merge expects a tributary')

    def _merge(foo1, foo2):
        for gen1, gen2 in zip(foo1(), foo2()):
            if isinstance(gen1, types.GeneratorType) and \
               isinstance(gen2, types.GeneratorType):
                for f1, f2 in zip(gen1, gen2):
                    ret = {}
                    ret.update(f1)
                    ret.update(f1)
                    yield ret
            elif isinstance(gen1, types.GeneratorType):
                for f1 in gen1:
                    ret = {}
                    ret.update(f1)
                    ret.update(gen2)
                    yield ret
            elif isinstance(gen2, types.GeneratorType):
                for f2 in gen2:
                    ret = {}
                    ret.update(gen1)
                    ret.update(f2)
                    yield ret
            else:
                ret = {}
                ret.update(gen1)
                ret.update(gen2)
                yield ret

    return _wrap(_merge, dict(foo1=f_wrap1, foo2=f_wrap2), name='DictMerge', wraps=(f_wrap1, f_wrap2), share=None)


def Reduce(*f_wraps):
    for f_wrap in f_wraps:
        if not isinstance(f_wrap, FunctionWrapper):
            raise Exception('Reduce expects a tributary')

    def _reduce(foos):
        for all_gens in zip(*[foo() for foo in foos]):
            gens = []
            vals = []
            for gen in all_gens:
                if isinstance(gen, types.GeneratorType):
                    gens.append(gen)
                else:
                    vals.append(gen)
            if gens:
                for gens in zip(*gens):
                    ret = list(vals)
                    for gen in gens:
                        ret.append(next(gen))
                    yield ret
            else:
                yield vals

    return _wrap(_reduce, dict(foos=f_wraps), name='Reduce', wraps=tuple(f_wraps), share=None)



import time
import math
import numpy as np

from ...thread import run  # noqa: F401
from ...base import StreamNone, StreamEnd  # noqa: F401
from ..base import _wrap, FunctionWrapper  # noqa: F401
from .ws import WebSocket, SyncWebSocket  # noqa: F401
from .http import HTTP, SyncHTTP  # noqa: F401
from .socketio import SocketIO, SyncSocketIO  # noqa: F401
from .kafka import Kafka, SyncKafka  # noqa: F401


def _gen():
    S = 100
    T = 252
    mu = 0.25
    vol = 0.5

    returns = np.random.normal(mu / T, vol / math.sqrt(T), T) + 1
    _list = returns.cumprod() * S
    return _list


def Random(size=10, interval=0.1):

    def _random(size, interval):
        step = 0
        while step < size:
            x = {y: _gen() for y in ('A', 'B', 'C', 'D')}
            for i in range(len(x['A'])):
                if step >= size:
                    break
                yield {'A': x['A'][i],
                       'B': x['B'][i],
                       'C': x['C'][i],
                       'D': x['D'][i]}
                time.sleep(interval)
                step += 1

    return _wrap(_random, dict(size=size, interval=interval), name='Random')



import types
import sys
from six import iteritems


def _wrap(foo, foo_kwargs, name='', wraps=(), share=None, state=None):
    if isinstance(foo, FunctionWrapper):
        ret = foo
    else:
        ret = FunctionWrapper(foo, foo_kwargs, name, wraps, share, state)

    for wrap in wraps:
        if isinstance(wrap, FunctionWrapper):
            if wrap == foo:
                continue
            _inc_ref(wrap, ret)
    return ret


def _call_if_function(f):
    if isinstance(f, types.FunctionType):
        return f()
    return f


def _inc_ref(f_wrapped, f_wrapping):
    if f_wrapped == f_wrapping:
        raise Exception('Internal Error')

    if f_wrapped._using is None or f_wrapped._using == id(f_wrapping):
        f_wrapped._using = id(f_wrapping)
        return
    Share(f_wrapped)
    f_wrapped._using = id(f_wrapping)


class FunctionWrapper(object):
    def __init__(self, foo, foo_kwargs, name='', wraps=(), share=None, state=None):
        state = state or {}

        if len(foo.__code__.co_varnames) > 0 and \
           foo.__code__.co_varnames[0] == 'state':
            self._foo = foo.__get__(self, FunctionWrapper)
            for k, v in iteritems(state):
                if k not in ('_foo', '_foo_kwargs', '_refs_orig', '_name', '_wraps', '_share'):
                    setattr(self, k, v)
                else:
                    raise Exception('Reserved Word - %s' % k)
        else:
            self._foo = foo

        self._foo_kwargs = foo_kwargs
        self._refs_orig, self._refs = 1, 1

        self._name = name
        self._wraps = wraps
        self._using = None
        self._share = share if share else self

    def get_last(self):
        if not hasattr(self, '_last'):
            raise Exception('Never called!!')

        if self._refs < 0:
            raise Exception('Ref mismatch in %s' % str(self._foo))

        self._refs -= 1
        return self._last

    def set_last(self, val):
        self._refs = self._refs_orig
        self._last = val

    last = property(get_last, set_last)

    def inc(self):
        self._refs_orig += 1
        self._refs += 1

    def view(self, _id=0, _idmap=None):
        _idmap = _idmap or {}
        ret = {}

        # check if i exist already in graph
        if id(self) in _idmap:
            key = _idmap[id(self)]
        else:
            key = self._name + str(_id)
            # _id += 1
            _idmap[id(self)] = key
        _id += 1

        ret[key] = []
        for f in self._wraps:
            if isinstance(f, FunctionWrapper):
                r, m = f.view(_id, _idmap)
                ret[key].append(r)
                _id = m
            else:
                if 'pandas' in sys.modules:
                    import pandas as pd
                    if isinstance(f, pd.DataFrame) or isinstance(f, pd.Series):
                        # pprint
                        f = 'DataFrame'
                ret[key].append(str(f))
        return ret, _id

    def __call__(self, *args, **kwargs):
        while(self._refs == self._refs_orig):
            kwargs.update(self._foo_kwargs)
            ret = self._foo(*args, **kwargs)
            if isinstance(ret, types.GeneratorType):
                for r in ret:
                    tmp = _call_if_function(r)

                    if isinstance(tmp, types.GeneratorType):
                        for rr in tmp:
                            self.last = rr
                            yield self.last

                    else:
                        self.last = tmp
                        yield self.last
            else:
                tmp = _call_if_function(ret)

                if isinstance(tmp, types.GeneratorType):
                    for rr in tmp:
                        self.last = rr
                        yield self.last

                else:
                    self.last = tmp
                    yield self.last
        while(0 < self._refs < self._refs_orig):
            yield self.last

        # reset state to be called again
        self._refs = self._refs_orig

    def __iter__(self):
        c_gen = self.__call__()
        for c in c_gen:
            yield c

    def __add__(self, other):
        pass


def Const(val):
    def _always(val):
        yield val

    return _wrap(_always, dict(val=val), name='Const', wraps=(val,))


def Foo(foo, foo_kwargs=None):
    return _wrap(foo, foo_kwargs or {}, name='Foo', wraps=(foo,))


def Share(f_wrap):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('Share expects a tributary')
    f_wrap.inc()
    return f_wrap



import ujson
from confluent_kafka import Consumer, KafkaError
from ..base import _wrap


def Kafka(servers, group, topics, json=False, wrap=False, interval=1):
    return AsyncKafka(servers, group, topics, json=json, wrap=wrap, interval=interval)


def AsyncKafka(servers, group, topics, json=False, wrap=False, interval=1):
    c = Consumer({
        'bootstrap.servers': servers,
        'group.id': group,
        'default.topic.config': {
            'auto.offset.reset': 'smallest'
        }
    })

    if not isinstance(topics, list):
        topics = [topics]
    c.subscribe(topics)

    async def _listen(consumer, json, wrap, interval):
        while True:
            msg = consumer.poll(interval)

            if msg is None:
                continue
            if msg.error():
                if msg.error().code() == KafkaError._PARTITION_EOF:
                    continue
                else:
                    print(msg.error())
                    break

            msg = msg.value().decode('utf-8')

            if not msg:
                break
            if json:
                msg = ujson.loads(msg)
            if wrap:
                msg = [msg]
            yield msg

    return _wrap(_listen, dict(consumer=c, json=json, wrap=wrap, interval=interval), name='Kafka')



import six
import functools
import inspect


class Node(object):
    def __init__(self,
                 name,
                 derived=False,
                 readonly=False,
                 nullable=False,
                 default_or_starting_value=None,
                 callable=None,
                 callable_args=None,
                 callable_kwargs=None,
                 callable_is_method=False,
                 always_dirty=False,
                 trace=False,
                 ):
        self._callable = callable
        self._name = name
        self._value = default_or_starting_value
        self._readonly = readonly
        self._trace = trace
        self._callable = callable
        self._callable_args = self._transform_args(callable_args or [])
        self._callable_kwargs = self._transform_kwargs(callable_kwargs or {})
        self._callable_is_method = callable_is_method
        self._always_dirty = always_dirty

        self._self_reference = self

        if callable:
            self._callable._node_wrapper = None  # not known until program start
            self._dependencies = {self._callable: (self._callable_args, self._callable_kwargs)}
        else:
            self._dependencies = {}

        if derived:
            self._dirty = True
        else:
            self._dirty = False

    def _transform_args(self, args):
        return args

    def _transform_kwargs(self, kwargs):
        return kwargs

    def _with_self(self, other_self):
        self._self_reference = other_self
        return self

    def _compute_from_dependencies(self):
        if self._dependencies:
            for deps in six.itervalues(self._dependencies):
                # recompute args
                for arg in deps[0]:
                    arg._recompute()

                # recompute kwargs
                for kwarg in six.itervalues(deps[1]):
                    kwarg._recompute()

            k = list(self._dependencies.keys())[0]

            if self._callable_is_method:
                new_value = k(self._self_reference, *self._dependencies[k][0], **self._dependencies[k][1])
            else:
                new_value = k(*self._dependencies[k][0], **self._dependencies[k][1])

            if isinstance(new_value, Node):
                k._node_wrapper = new_value
                new_value = new_value()  # get value

            if self._trace:
                if new_value != self._value:
                    print('recomputing: %s-%d' % (self._name, id(self)))
            self._value = new_value
        return self._value

    def _subtree_dirty(self):
        for deps in six.itervalues(self._dependencies):
            # check args
            for arg in deps[0]:
                if arg._dirty or arg._subtree_dirty() or arg._always_dirty:
                    return True

            # check kwargs
            for kwarg in six.itervalues(deps[1]):
                if kwarg._dirty or kwarg._subtree_dirty() or kwarg._always_dirty:
                    return True
        return False

    def _recompute(self):
        self._dirty = self._dirty or self._subtree_dirty() or self._always_dirty
        if self._dirty:
            self._value = self._compute_from_dependencies()
        self._dirty = False

    def _gennode(self, other, name, foo, foo_args):
        return Node(name='{name}_{lhs}_{rhs}'.format(name=name, lhs=self._name, rhs=other._name),
                    derived=True,
                    callable=foo,
                    callable_args=foo_args,
                    trace=self._trace or other._trace)

    def _tonode(self, other):
        if isinstance(other, Node):
            return other
        return Node(name='gen_' + str(other)[:5],
                    derived=True,
                    default_or_starting_value=other,
                    trace=self._trace)

    def set(self, **kwargs):
        for k, v in six.iteritems(kwargs):
            _set = False
            for deps in six.itervalues(self._dependencies):
                # try to set args
                for arg in deps[0]:
                    if arg._name == k:
                        arg._dirty = arg._value != v
                        arg._value = v
                        _set = True
                        break

                if _set:
                    continue

                # try to set kwargs
                for kwarg in six.itervalues(deps[1]):
                    if kwarg._name == k:
                        kwarg._dirty = kwarg._value != v
                        kwarg._value = v
                        _set = True
                        break

    def value(self):
        return self._value

    def print(self, counter=0):
        key = str(self) + '#' + str(counter)
        ret = {key: []}
        counter += 1
        if self._dependencies:
            for call, deps in six.iteritems(self._dependencies):
                # callable node
                if hasattr(call, '_node_wrapper') and \
                   call._node_wrapper is not None:
                    ret[key].append(call._node_wrapper.print(counter))
                    counter += 1

                # args
                for arg in deps[0]:
                    ret[key].append(arg.print(counter))
                    counter += 1

                # kwargs
                for kwarg in six.itervalues(deps[1]):
                    ret[key].append(kwarg.print(counter))
                    counter += 1

        return ret

    def graph(self):
        return self.print()

    def graphviz(self):
        d = self.graph()
        from graphviz import Digraph
        dot = Digraph(self._name, strict=True)
        dot.format = 'png'

        def rec(nodes, parent):
            for d in nodes:
                if not isinstance(d, dict):
                    dot.node(d)
                    dot.edge(d, parent)

                else:
                    for k in d:
                        dot.node(k)
                        rec(d[k], k)
                        dot.edge(k, parent)

        for k in d:
            dot.node(k)
            rec(d[k], k)
        return dot

    def __call__(self):
        self._recompute()
        return self.value()

    def __add__(self, other):
        other = self._tonode(other)
        if isinstance(self._self_reference, Node):
            return self._gennode(other, '+', (lambda x, y: x.value() + y.value()), [self._self_reference, other])
        return self._gennode(other, '+', (lambda x, y: x.value() + y.value()), [self, other])

    def __sub__(self, other):
        other = self._tonode(other)
        if isinstance(self._self_reference, Node):
            return self._gennode(other, '-', (lambda x, y: x.value() - y.value()), [self._self_reference, other])
        return self._gennode(other, '-', (lambda x, y: x.value() - y.value()), [self, other])

    def __bool__(self):
        return self.value()

    __nonzero__ = __bool__  # Py2 compat

    def __eq__(self, other):
        if isinstance(other, Node) and super(Node, self).__eq__(other):
            return True

        other = self._tonode(other)
        return self._gennode(other, '==', (lambda x, y: x() == y()), [self._self_reference, other])

    def __ne__(self, other):
        if isinstance(other, Node) and super(Node, self).__eq__(other):
            return False

        other = self._tonode(other)
        return self._gennode(other, '!=', (lambda x, y: x() != y()), [self, other])

    def __ge__(self, other):
        if isinstance(other, Node) and super(Node, self).__eq__(other):
            return True

        other = self._tonode(other)
        return self._gennode(other, '>=', (lambda x, y: x() >= y()), [self, other])

    def __gt__(self, other):
        if isinstance(other, Node) and super(Node, self).__eq__(other):
            return False

        other = self._tonode(other)
        return self._gennode(other, '>', (lambda x, y: x() > y()), [self, other])

    def __le__(self, other):
        if isinstance(other, Node) and super(Node, self).__eq__(other):
            return True

        other = self._tonode(other)
        return self._gennode(other, '<=', (lambda x, y: x() <= y()), [self, other])

    def __lt__(self, other):
        if isinstance(other, Node) and super(Node, self).__eq__(other):
            return False
        other = self._tonode(other)
        return self._gennode(other, '<', (lambda x, y: x() < y()), [self, other])

    def __repr__(self):
        return '%s' % (self._name)


class BaseClass(object):
    def __init__(self, *args, **kwargs):
        pass

    def node(self, name, readonly=False, nullable=True, default_or_starting_value=None, trace=False):
        if not hasattr(self, '_BaseClass__nodes'):
            self.__nodes = {}

        self.__nodes[name] = Node(name=name,
                                  derived=False,
                                  readonly=readonly,
                                  nullable=nullable,
                                  default_or_starting_value=default_or_starting_value,
                                  trace=trace)
        setattr(self, name, self.__nodes[name])

        return self.__nodes[name]

    def __getattribute__(self, name):
        if name == '_BaseClass__nodes' or name == '__nodes':
            return super(BaseClass, self).__getattribute__(name)
        elif hasattr(self, '_BaseClass__nodes') and name in super(BaseClass, self).__getattribute__('_BaseClass__nodes'):
            # return super(BaseClass, self).__getattribute__('_BaseClass__nodes')[name]._value
            return super(BaseClass, self).__getattribute__('_BaseClass__nodes')[name]
        else:
            return super(BaseClass, self).__getattribute__(name)

    def __setattr__(self, name, value):
        if hasattr(self, '_BaseClass__nodes') and name in super(BaseClass, self).__getattribute__('_BaseClass__nodes'):
            node = super(BaseClass, self).__getattribute__('_BaseClass__nodes')[name]
            if isinstance(value, Node) and node == value:
                return
            elif isinstance(value, Node):
                raise Exception('Cannot set to node')
            else:
                node._dirty = node._value != value
                node._value = value
        else:
            super(BaseClass, self).__setattr__(name, value)


def _either_type(f):
    @functools.wraps(f)
    def new_dec(*args, **kwargs):
        if len(args) == 1 and len(kwargs) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwargs)
    return new_dec


@_either_type
def node(meth, memoize=False, trace=False):
    argspec = inspect.getfullargspec(meth)
    # args = argspec.args
    # varargs = argspec.varargs
    # varkw = argspec.varkw
    # defaults = argspec.defaults
    # kwonlyargs = argspec.kwonlyargs
    # kwonlydefaults = args.kwonlydefaults

    if argspec.varargs:
        raise Exception('varargs not supported yet!')
    if argspec.varkw:
        raise Exception('varargs not supported yet!')

    node_args = []
    node_kwargs = {}
    is_method = False

    for i, arg in enumerate(argspec.args):
        if arg == 'self':
            # TODO
            is_method = True
            continue

        if (is_method and len(argspec.defaults or []) >= i) or \
           (not is_method and len(argspec.defaults or []) > i):
            default_or_starting_value = argspec.defaults[0]
            nullable = True
        else:
            default_or_starting_value = None
            nullable = False

        node_args.append(Node(name=arg,
                              derived=True,
                              readonly=False,
                              nullable=nullable,
                              default_or_starting_value=default_or_starting_value,
                              trace=trace))

    for k, v in six.iteritems(argspec.kwonlydefaults or {}):
        node_kwargs[k] = Node(name=k,
                              derived=True,
                              readonly=False,
                              nullable=True,
                              default_or_starting_value=v,
                              trace=trace)

    def meth_wrapper(self, *args, **kwargs):
        if len(args) > len(node_args):
            raise Exception('Missing args (call or preprocessing error has occurred)')

        if len(kwargs) > len(node_kwargs):
            raise Exception('Missing kwargs (call or preprocessing error has occurred)')

        # move to __call__
        # # update args
        # for i, value in args:
        #     node = node_args[i]
        #     if node._value != value:
        #         node._dirty = True
        #     node._value = value

        # # update kwargs
        # for k, value in six.iteritems(kwargs):
        #     node = node_kwargs[k]
        #     if node._value != value:
        #         node._dirty = True
        #     node._value = value

        if is_method:
            val = meth(self, *(arg.value() for arg in args), **kwargs)
        else:
            val = meth(*(arg.value() for arg in args), **kwargs)
        return val

    new_node = Node(name=meth.__name__,
                    derived=True,
                    callable=meth_wrapper,
                    callable_args=node_args,
                    callable_kwargs=node_kwargs,
                    callable_is_method=is_method,
                    always_dirty=not memoize,
                    trace=trace)

    if is_method:
        ret = lambda self, *args, **kwargs: new_node._with_self(self)  # noqa: E731
    else:
        ret = lambda *args, **kwargs: new_node  # noqa: E731

    ret._node_wrapper = new_node
    return ret



import requests
import ujson
from ..base import _wrap


def HTTP(foo, foo_kwargs=None, *args, **kwargs):
    return SyncHTTP(foo, foo_kwargs, *args, **kwargs)


def SyncHTTP(foo, foo_kwargs=None, url='', json=False, wrap=False, field=None, proxies=None, cookies=None):
    foo_kwargs = foo_kwargs or {}
    foo = _wrap(foo, foo_kwargs)

    def _send(foo, url, json=False, wrap=False, field=None, proxies=None, cookies=None):
        for data in foo():
            if wrap:
                data = [data]
            if json:
                data = ujson.dumps(data)

            msg = requests.post(url, data=data, cookies=cookies, proxies=proxies)

            if msg is None or msg.status_code != 200:
                break

            if json:
                msg = msg.json()

            if field:
                msg = msg[field]

            if wrap:
                msg = [msg]

            yield msg

    return _wrap(_send, dict(foo=foo, url=url, json=json, wrap=wrap, field=field, proxies=proxies, cookies=cookies), name='HTTP')



import types
from ..base import _wrap


def Count(foo, foo_kwargs=None):
    foo_kwargs = foo_kwargs or {}
    foo = _wrap(foo, foo_kwargs)

    def _count(foo):
        count = 0
        for gen in foo():
            if isinstance(gen, types.GeneratorType):
                for f in gen:
                    count += 1
                    yield count
            else:
                count += 1
                yield count

    return _wrap(_count, dict(foo=foo), name='Count', wraps=(foo,), share=foo)


def Sum(foo, foo_kwargs=None):
    foo_kwargs = foo_kwargs or {}
    foo = _wrap(foo, foo_kwargs)

    def _sum(foo):
        sum = 0
        for gen in foo():
            if isinstance(gen, types.GeneratorType):
                for f in gen:
                    sum += f
                    yield sum
            else:
                sum += gen
                yield sum

    return _wrap(_sum, dict(foo=foo), name='Sum', wraps=(foo,), share=foo)



import asyncio
import time
import types
from .base import _wrap, FunctionWrapper, Foo, Const


def Timer(foo_or_val, kwargs=None, interval=1, repeat=0):
    kwargs = kwargs or {}

    if not isinstance(foo_or_val, types.FunctionType):
        foo = Const(foo_or_val)
    else:
        foo = Foo(foo_or_val, kwargs)

    async def _repeater(foo, repeat, interval):
        while repeat > 0:
            t1 = time.time()
            f = foo()
            yield f
            t2 = time.time()

            if interval > 0:
                # sleep for rest of time that _p didnt take
                await asyncio.sleep(max(0, interval-(t2-t1)))
            repeat -= 1

    return _wrap(_repeater, dict(foo=foo, repeat=repeat, interval=interval), name='Timer', wraps=(foo,), share=foo)


def State(foo, foo_kwargs=None, **state):
    foo_kwargs = foo_kwargs or {}
    foo = _wrap(foo, foo_kwargs, name=foo.__name__, wraps=(foo,), state=state)
    return foo


def Apply(foo, f_wrap, foo_kwargs=None):
    if not isinstance(f_wrap, FunctionWrapper):
        raise Exception('Apply expects a tributary')
    foo_kwargs = foo_kwargs or {}
    foo = Foo(foo, foo_kwargs)
    foo._wraps = foo._wraps + (f_wrap, )

    async def _apply(foo):
        async for f in f_wrap():
            yield foo(f)

    return _wrap(_apply, dict(foo=foo), name='Apply', wraps=(foo,), share=foo)


def Window(foo, foo_kwargs=None, size=-1, full_only=True):
    foo_kwargs = foo_kwargs or {}
    foo = Foo(foo, foo_kwargs)

    accum = []

    async def _window(foo, size, full_only, accum):
        async for x in foo():
            if size == 0:
                yield x
            else:
                accum.append(x)

                if size > 0:
                    accum = accum[-size:]
                if full_only:
                    if len(accum) == size or size == -1:
                        yield accum
                else:
                    yield accum

    return _wrap(_window, dict(foo=foo, size=size, full_only=full_only, accum=accum), name='Window', wraps=(foo,), share=foo)



import sys
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

if sys.version_info.major < 3:
    with open(path.join(here, 'requirements_py2.txt'), encoding='utf-8') as f:
        requires = f.read().split()
else:
    with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
        requires = f.read().split()


setup(
    name='tributary',
    version='0.0.4',
    description='Analytics library',
    long_description=long_description,
    url='https://github.com/timkpaine/tributary',
    download_url='https://github.com/timkpaine/tributary/archive/v0.0.4.tar.gz',
    author='Tim Paine',
    author_email='timothy.k.paine@gmail.com',
    license='Apache 2.0',
    install_requires=requires,
    extras_require={'dev': requires + ['pytest', 'pytest-cov', 'pylint', 'flake8']},

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    keywords='analytics tools plotting',

    packages=find_packages(exclude=['tests', ]),
    package_data={},
    include_package_data=True,
    zip_safe=False,
    entry_points={
    }
)



from gevent import monkey
_PATCHED = False
if not _PATCHED:
    monkey.patch_all(thread=False, select=False)
    _PATCHED = True


__version__ = '0.0.4'


from .functional import pipeline, stop  # noqa: F401
from .reactive import *  # noqa: F401, F403



from .base import _wrap, Foo, Const, Share  # noqa: F401
from .utils import *  # noqa: F401, F403
from .calculations import *  # noqa: F401, F403
from .input import *  # noqa: F401, F403
from .output import *  # noqa: F401, F403


def run(foo, **kwargs):
    foo = _wrap(foo, kwargs)
    ret = []
    try:
        for item in foo():
            ret.append(item)
    except KeyboardInterrupt:
        print('Terminating...')
    return ret



import types
from aiostream.stream import zip
from ..base import _wrap, FunctionWrapper


def unary(lam, foo, foo_kwargs=None, _name=''):
    foo_kwargs = None or {}
    foo = _wrap(foo, foo_kwargs)

    async def _unary(foo):
        async for gen in foo():
            if isinstance(gen, types.AsyncGeneratorType):
                async for f in gen:
                    if isinstance(f, types.CoroutineType):
                        yield lam(await f)
                    else:
                        yield lam(f)
            elif isinstance(gen, types.CoroutineType):
                yield lam(await gen)
            else:
                yield lam(gen)

    return _wrap(_unary, dict(foo=foo), name=_name or 'Unary', wraps=(foo,), share=None)


def bin(lam, foo1, foo2, foo1_kwargs=None, foo2_kwargs=None, _name=''):
    foo1_kwargs = None or {}
    foo2_kwargs = None or {}
    foo1 = _wrap(foo1, foo1_kwargs)
    foo2 = _wrap(foo2, foo2_kwargs)

    async def _bin(foo1, foo2):
        # TODO replace with merge
        async for gen1, gen2 in zip(foo1(), foo2()):
            if isinstance(gen1, types.AsyncGeneratorType) and isinstance(gen2, types.AsyncGeneratorType):
                print(1)
                async for f1, f2 in zip(gen1, gen2):
                    if isinstance(f1, types.CoroutineType):
                        f1 = await f1
                    if isinstance(f2, types.CoroutineType):
                        f2 = await f2
                    yield lam(f1, f2)
            elif isinstance(gen1, types.AsyncGeneratorType):
                async for f1 in gen1:
                    if isinstance(f1, types.CoroutineType):
                        f1 = await f1
                    if isinstance(gen2, types.CoroutineType):
                        gen2 = await gen2
                    yield lam(f1, gen2)
            elif isinstance(gen2, types.AsyncGeneratorType):
                print(3)
                async for f2 in gen2:
                    print(31)
                    if isinstance(gen1, types.CoroutineType):
                        gen1 = await gen1
                    if isinstance(f2, types.CoroutineType):
                        f2 = await f2

                    yield lam(gen1, f2)
            else:
                if isinstance(gen1, types.CoroutineType):
                    gen1 = await gen1
                if isinstance(gen2, types.CoroutineType):
                    gen2 = await gen2
                yield lam(gen1, gen2)

    return _wrap(_bin, dict(foo1=foo1, foo2=foo2), name=_name or 'Binary', wraps=(foo1, foo2), share=None)


def Noop(foo, foo_kwargs=None):
    return unary(lambda x: x, foo, foo_kwargs, _name='Noop')


def Negate(foo, foo_kwargs=None):
    return unary(lambda x: -1*x, foo, foo_kwargs, _name='Negate')


def Invert(foo, foo_kwargs=None):
    return unary(lambda x: 1/x, foo, foo_kwargs, _name='Invert')


def Not(foo, foo_kwargs=None):
    return unary(lambda x: not x, foo, foo_kwargs, _name='Not')


def Add(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x+y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Add')


def Sub(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x-y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Sub')


def Mult(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x*y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Mult')


def Div(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x/y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Div')


def Mod(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x % y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Mod')


def Pow(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x**y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Pow')


def And(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x and y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='And')


def Or(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x or y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Or')


def Equal(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x == y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Equal')


def Less(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x < y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='Less')


def More(foo1, foo2, foo1_kwargs=None, foo2_kwargs=None):
    return bin(lambda x, y: x > y, foo1, foo2, foo1_kwargs, foo2_kwargs, _name='More')


# Arithmetic
FunctionWrapper.__add__ = Add
FunctionWrapper.__sub__ = Sub
FunctionWrapper.__mul__ = Mult
FunctionWrapper.__div__ = Div
FunctionWrapper.__truediv__ = Div
FunctionWrapper.__pow__ = Pow
FunctionWrapper.__mod__ = Mod

# Logical
# FunctionWrapper.__and__ = And
# FunctionWrapper.__or__ = Or
# FunctionWrapper.__invert__ = Not
# TODO use __bool__ operator

# Comparator
FunctionWrapper.__lt__ = Less
FunctionWrapper.__gt__ = More
